/*
 * Copyright 2002 Sun Microsystems, Inc. All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or
 * without modification, are permitted provided that the following
 * conditions are met:
 * 
 * - Redistributions of source code must retain the above copyright
 *   notice, this list of conditions and the following disclaimer.
 * 
 * - Redistribution in binary form must reproduce the above
 *   copyright notice, this list of conditions and the following
 *   disclaimer in the documentation and/or other materials
 *   provided with the distribution.
 * 
 * Neither the name of Sun Microsystems, Inc. or the names of
 * contributors may be used to endorse or promote products derived
 * from this software without specific prior written permission.
 * 
 * This software is provided "AS IS," without a warranty of any
 * kind. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE OR NON-INFRINGEMENT, ARE HEREBY
 * EXCLUDED. SUN AND ITS LICENSORS SHALL NOT BE LIABLE FOR ANY
 * DAMAGES OR LIABILITIES SUFFERED BY LICENSEE AS A RESULT OF OR
 * RELATING TO USE, MODIFICATION OR DISTRIBUTION OF THIS SOFTWARE OR
 * ITS DERIVATIVES. IN NO EVENT WILL SUN OR ITS LICENSORS BE LIABLE
 * FOR ANY LOST REVENUE, PROFIT OR DATA, OR FOR DIRECT, INDIRECT,
 * SPECIAL, CONSEQUENTIAL, INCIDENTAL OR PUNITIVE DAMAGES, HOWEVER
 * CAUSED AND REGARDLESS OF THE THEORY OF LIABILITY, ARISING OUT OF
 * THE USE OF OR INABILITY TO USE THIS SOFTWARE, EVEN IF SUN HAS
 * BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
 * 
 * You acknowledge that this software is not designed, licensed or
 * intended for use in the design, construction, operation or
 * maintenance of any nuclear facility.
 */

package com.sun.java.swing.action;

import javax.swing.Action;
import javax.swing.ImageIcon;

import java.util.HashMap;

/**
 * The ActionManager manages and coordinates the enabling and handling of actions 
 * between all the UI elements like toolbars, menus and popup menus.
 * Acts as the glue (controller) between the UI and the model.
 *
 * @version  %I% %G%
 * @author  Mark Davidson
 */
public abstract class ActionManager {
    
    // Table of all possible actions.
    // key: CMD_NAME_...    value: instanceof AbstractAction
    private HashMap actions;
    
    private static ActionUtilities utilities = new ActionUtilities();

    /**
     * Shared instance of the singleton ActionManager.
     */
    protected static ActionManager manager;

    /**
     * Creates the action manager
     */
    protected ActionManager() {
        actions = new HashMap();
	addActions();
    }
    
    /**
     * Return the shared instance of the ActionManger. Subclassese should
     * override this method to construct the ActionManager the first time
     * this method is called.
     * <p>
     * Note: Unfortunately the JLS doesn't permit abstract and static modifiers
     */
    public static ActionManager getInstance() {
        return manager;
    }

    /** 
     * Configures the action manager to manage specific actions.
     * Subclasses should implement this method to add application
     * specific actions.
     */
    protected abstract void addActions();

    /** 
     * Adds an action to the ActionManager
     */
    protected void addAction(String cmdname, Action action)  {
        actions.put(cmdname, action);
    }


    
    /** 
     * Retreives the action for an widget that requires Action objects.
     * Such as JMenu and JToolBar.
     * @param key One of ActionConstants.CMD_NAME_...
     */
    public Action getAction(String key)  {
	return (Action)actions.get(key);
    }
    
    /**
     * Convenience method for returning the DelegateAction
     *
     * @return the DelegateAction referenced by the named key or null
     */
    public DelegateAction getDelegateAction(String name) {
	Action a = getAction(name);
	if (a instanceof DelegateAction) {
	    return (DelegateAction)a;
	}
	return null;
    }

    /**
     * Convenience method for returning the StateChangeAction
     *
     * @return the StateChangeAction referenced by the named key or null
     */
    public StateChangeAction getStateChangeAction(String name) {
	Action a = getAction(name);
	if (a instanceof StateChangeAction) {
	    return (StateChangeAction)a;
	}
	return null;
    }

    /**
     * Returns the Icon associated with the name of the resource.
     */
    public static ImageIcon getIcon(String name) {
	return utilities.getIcon(name);
    }

    /**
     * Enables or disables the state of the action
     *
     * @param name Name of the action (key)
     * @param enabled true if the action is to be enabled; otherwise false
     */
    public void setActionEnabled(String name, boolean enabled) {
	Action action = getAction(name);
	if (action != null) {
	    action.setEnabled(enabled);
	}
    }

}

class ActionUtilities {

    // Image directories URL for JLF graphics
    public static final String IMAGE_DIR = "/toolbarButtonGraphics/";

    public ImageIcon getIcon(String name) {
	String imagePath = IMAGE_DIR + name;
	java.net.URL url = this.getClass().getResource(imagePath);
	if (url != null) {
	    return new ImageIcon(url);
	}
	return null;
    }
}
