/*
 * Copyright 2002 Sun Microsystems, Inc. All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or
 * without modification, are permitted provided that the following
 * conditions are met:
 * 
 * - Redistributions of source code must retain the above copyright
 *   notice, this list of conditions and the following disclaimer.
 * 
 * - Redistribution in binary form must reproduce the above
 *   copyright notice, this list of conditions and the following
 *   disclaimer in the documentation and/or other materials
 *   provided with the distribution.
 * 
 * Neither the name of Sun Microsystems, Inc. or the names of
 * contributors may be used to endorse or promote products derived
 * from this software without specific prior written permission.
 * 
 * This software is provided "AS IS," without a warranty of any
 * kind. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE OR NON-INFRINGEMENT, ARE HEREBY
 * EXCLUDED. SUN AND ITS LICENSORS SHALL NOT BE LIABLE FOR ANY
 * DAMAGES OR LIABILITIES SUFFERED BY LICENSEE AS A RESULT OF OR
 * RELATING TO USE, MODIFICATION OR DISTRIBUTION OF THIS SOFTWARE OR
 * ITS DERIVATIVES. IN NO EVENT WILL SUN OR ITS LICENSORS BE LIABLE
 * FOR ANY LOST REVENUE, PROFIT OR DATA, OR FOR DIRECT, INDIRECT,
 * SPECIAL, CONSEQUENTIAL, INCIDENTAL OR PUNITIVE DAMAGES, HOWEVER
 * CAUSED AND REGARDLESS OF THE THEORY OF LIABILITY, ARISING OUT OF
 * THE USE OF OR INABILITY TO USE THIS SOFTWARE, EVEN IF SUN HAS
 * BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
 * 
 * You acknowledge that this software is not designed, licensed or
 * intended for use in the design, construction, operation or
 * maintenance of any nuclear facility.
 */
package com.sun.java.swing.ui;

import javax.swing.Action;
import javax.swing.ButtonGroup;
import javax.swing.JCheckBoxMenuItem;
import javax.swing.JRadioButtonMenuItem;
import javax.swing.JMenu;
import javax.swing.JMenuBar;
import javax.swing.JMenuItem;

import com.sun.java.swing.action.ActionManager;
import com.sun.java.swing.action.StateChangeAction;

/**
 * Common menu bar based on Actions
 *
 * To add a new Action/command:
 * 1. Create a DelegateAction subclass.
 * 2. Add the Action to the ActionManager.
 * 3. Add the action from the ActionManager to the configureMenu method.
 *
 * @version  %I% %G%
 * @author  Mark Davidson
 */
public abstract class CommonMenuBar extends JMenuBar {
    
    protected ActionManager manager;

    protected CommonMenuBar(ActionManager manager) {
	this.manager = manager;
        configureMenu();
    }

    /** 
     * Subclasses must implement this method to configure their menu.
     */
    protected abstract void configureMenu();
    
    
    /** 
     * override for custom configuration
     */
    protected void configureToggleMenuItem(JMenuItem menuItem, Action action) {
	configureMenuItem(menuItem, action);

	// The PropertyChangeListener that gets added
	// to the Action doesn't know how to handle the "selected" property change
	// in the meantime, the corect thing to do is to add another PropertyChangeListener
	// to the StateChangeAction until this is fixed.
	action.addPropertyChangeListener(new ToggleActionPropertyChangeListener(menuItem));
    }


    /** 
     * override for custom configuration
     */
    protected void configureMenuItem(JMenuItem menuItem, Action action) {
        // The name is the key into the ActionManager action list.
        menuItem.setName((String)action.getValue(Action.NAME));
    }

    /** 
     * Helper method to ensure that menu characteristics are consistent.
     * @param name Name of the menu
     * @param mnemonic Accelerator 
     */    
    protected JMenu createMenu(String name, char mnemonic)  {
        JMenu menu = new JMenu(name);
        menu.setMnemonic(mnemonic);
        
        return menu;
    }
    
    /** 
     * Helper method to configure each item consistantly
     */
    protected void addMenuItem(JMenu menu, Action action)  {
        JMenuItem menuItem = menu.add(action);
        
        configureMenuItem(menuItem, action);
    }
    
    /** 
     * Helper method to add a checkbox menu item
     */
    protected void addCheckBoxMenuItem(JMenu menu, StateChangeAction a)  {
        addCheckBoxMenuItem(menu, a, false);
    }

    /** 
     * Helper method to add a checkbox menu item.
     */
    protected void addCheckBoxMenuItem(JMenu menu, StateChangeAction a,
                            boolean selected)  {
        JCheckBoxMenuItem mi = new JCheckBoxMenuItem(a);
        mi.addItemListener(a);
        mi.setSelected(selected);
        menu.add(mi);

        configureToggleMenuItem(mi, a);
    }
    
    /** 
     * Helper method to add a radio button menu item.
     */
    protected void addRadioButtonMenuItem(JMenu menu, ButtonGroup group, Action a)  {
       addRadioButtonMenuItem(menu, group, a, false);
    }
    
    /** 
     * Helper method to add a radio button menu item.
     */
    protected void addRadioButtonMenuItem(JMenu menu, ButtonGroup group, 
                        Action a, boolean selected)  {
        JRadioButtonMenuItem mi = new JRadioButtonMenuItem(a);
        mi.setSelected(selected);
        menu.add(mi);
        group.add(mi);

        configureToggleMenuItem(mi, a);
    }
    
}
